package com.fzm.common.service.impl;

import cn.hutool.core.util.IdUtil;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fzm.common.entity.Order;
import com.fzm.common.entity.Payment;
import com.fzm.common.entity.dto.JsapiPayDto;
import com.fzm.common.entity.dto.OrderProcessMsg;
import com.fzm.common.enums.*;
import com.fzm.common.exception.GlobalException;
import com.fzm.common.properties.WxPayProperties;
import com.fzm.common.service.*;
import com.fzm.common.utils.SnowflakeUtil;
import com.wechat.pay.contrib.apache.httpclient.util.AesUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.util.EntityUtils;
import org.redisson.Redisson;
import org.redisson.api.RLock;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Base64Utils;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.PrivateKey;
import java.security.Signature;
import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * @author tangtuo
 * @date 2022/1/13 10:13
 */
@Slf4j
@Service
@Transactional(rollbackFor = RuntimeException.class)
public class WxPayServiceImpl implements WxPayService {


    @Resource
    private CloseableHttpClient httpClient;

    @Resource
    private WxPayProperties wxPayProperties;

    @Resource
    private PrivateKey wxPrivateKey;

    @Resource
    private OrderService orderService;

    @Resource
    private Redisson redisson;

    @Resource
    private PaymentService paymentService;

    @Resource
    private RabbitTemplate rabbitTemplate;

    @Resource
    private SnowflakeUtil snowflakeUtil;

    public Map<String, Object> payJsapi(JsapiPayDto jsapiPayDto) throws Exception {
        RLock lock = redisson.getLock("pay-" + jsapiPayDto.getOrderId());
        if (!lock.tryLock(10, TimeUnit.SECONDS)) {
            throw GlobalException.newException(ResultCode.PAY_FAILED, "当前订单正在支付中,请勿重复点击");
        }
        try {
            String appId = wxPayProperties.getAppId();
            Long out_trade_no = jsapiPayDto.getOrderId();
            Order order = orderService.getById(out_trade_no);
            // 请求jsapi支付下单接口
            String url = "https://api.mch.weixin.qq.com/v3/pay/transactions/jsapi";
            HttpPost httpPost = new HttpPost(url);
            httpPost.addHeader("Accept", "application/json");
            httpPost.addHeader("Content-type", "application/json; charset=utf-8");
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectMapper objectMapper = new ObjectMapper();
            ObjectNode rootNode = objectMapper.createObjectNode();
            rootNode.put("mchid", wxPayProperties.getMchId())
                    .put("appid", appId)
                    .put("description", PayScene.getTypeByCode(order.getPayScene()))
                    .put("notify_url", wxPayProperties.getPayNotifyUrl())
                    .put("out_trade_no", String.valueOf(out_trade_no));
            rootNode.putObject("amount")
                    .put("total", order.getFee());
            rootNode.putObject("payer")
                    .put("openid", jsapiPayDto.getOpenid());
            String json = objectMapper.writeValueAsString(rootNode);
            log.info("h5下单接口请求参数: {}", json);
            objectMapper.writeValue(bos, rootNode);
            httpPost.setEntity(new StringEntity(bos.toString("UTF-8"), "UTF-8"));
            CloseableHttpResponse response = httpClient.execute(httpPost);
            String bodyAsString = EntityUtils.toString(response.getEntity());
            JSONObject jsonObject = JSONUtil.parseObj(bodyAsString);
            if (jsonObject == null || StringUtils.isBlank(jsonObject.getStr("prepay_id"))) {
                throw GlobalException.newException(ResultCode.PAY_FAILED, "微信支付下单失败");
            }
            String prepay_id = jsonObject.getStr("prepay_id");

            String timeStamp = String.valueOf(Instant.now().getEpochSecond());
            String nonceStr = IdUtil.simpleUUID();
            String packageInfo = "prepay_id=" + prepay_id;
            String signStr = appId + "\n" +
                    timeStamp + "\n" +
                    nonceStr + "\n" +
                    packageInfo + "\n";
            log.info("签名字符串为: {}", signStr);
            String paySign = this.encryptByPrivateKey(signStr);
            Map<String, Object> result = new HashMap<>();
            result.put("appId", appId);
            result.put("timeStamp", timeStamp);
            result.put("nonceStr", nonceStr);
            result.put("package", packageInfo);
            result.put("signType", "RSA");
            result.put("paySign", paySign);
            result.put("orderId", out_trade_no);
            // 修改此订单的支付方式为微信支付
            orderService.updatePayType(jsapiPayDto.getOrderId(), PayType.WXPAY.getCode());
            return result;
        } finally {
            lock.unlock();
        }
    }


    /**
     * 私钥签名
     *
     * @param data 需要加密的数据
     * @return 加密后的数据
     * @throws Exception 异常信息
     */
    public String encryptByPrivateKey(String data) throws Exception {
        Signature signature = Signature.getInstance("SHA256WithRSA");
        signature.initSign(wxPrivateKey);
        signature.update(data.getBytes(StandardCharsets.UTF_8));
        byte[] signed = signature.sign();
        return Base64Utils.encodeToString(signed);
    }

    @Override
    public Boolean notifyH5(HttpServletRequest request) {
        try {
            String requestBodyData = this.getRequestBodyData(request);
            log.info("接收微信支付回调请求, 请求参数: {}", requestBodyData);
            JSONObject jsonObject = JSONUtil.parseObj(requestBodyData);
            JSONObject resource = jsonObject.getJSONObject("resource");
            String plainText = this.decrypt(resource);
            log.info("解密后的明文信息为: {}", plainText);
            JSONObject obj = JSONUtil.parseObj(plainText);
            // 处理订单
            processOrder(obj);
        } catch (Exception e) {
            log.error("微信支付回调异常", e);
            return false;
        }
        return true;
    }


    @Override
    public void processOrder(JSONObject jsonObject) {
        // 获取订单号
        Long out_trade_no = jsonObject.getLong("out_trade_no");
        // 加锁,避免接口幂等性问题
        RLock lock = redisson.getLock("pay-notify-" + out_trade_no);
        lock.lock(10, TimeUnit.SECONDS);
        // 修改订单状态
        try {
            Order order = orderService.getById(out_trade_no);
            // 如果订单状态不是待支付,直接返回
            if (!OrderStatus.PAYING.getStatus().equals(order.getOrderStatus())) {
                log.warn("当前订单已处理完成, 订单号:==> {}", out_trade_no);
                return;
            }
            Integer productId = order.getProductId();
            // 订单支付成功后,给mq发送一条消息,处理nft发行或版权申请的状态
            OrderProcessMsg orderProcessMsg = new OrderProcessMsg(productId, order.getPayScene());
            rabbitTemplate.convertAndSend("order-exchange", "order.process", orderProcessMsg);
            // 修改订单状态
            orderService.updateOrderStatus(out_trade_no, OrderStatus.PAYED);
            // 插入支付流水记录
            Payment payment = new Payment();
            payment.setOrderId(out_trade_no);
            payment.setTradeType(jsonObject.getStr("trade_type"));
            payment.setTotalFee(jsonObject.getJSONObject("amount").getInt("payer_total"));
            payment.setTransactionId(jsonObject.getStr("transaction_id"));
            payment.setSuccessTime(jsonObject.getStr("success_time"));
            payment.setTradeState(jsonObject.getStr("trade_state"));
            payment.setContent(JSONUtil.toJsonStr(jsonObject));
            paymentService.save(payment);
        } finally {
            lock.unlock();
        }
    }


    @Override
    public String queryOrder(Long orderId) {

        return null;
    }

    @Override
    public void refund(Long orderId) throws IOException {
        Order order = orderService.getById(orderId);
        if (order == null || !order.getOrderStatus().equals(OrderStatus.PAYED.getStatus())) {
            throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单未支付成功");
        }
        String out_refund_no = "refund-" + snowflakeUtil.snowflakeId();
        String url = "https://api.mch.weixin.qq.com/v3/refund/domestic/refunds";
        HttpPost httpPost = new HttpPost(url);
        httpPost.addHeader("Accept", "application/json");
        httpPost.addHeader("Content-type", "application/json; charset=utf-8");
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectMapper objectMapper = new ObjectMapper();
        ObjectNode rootNode = objectMapper.createObjectNode();
        rootNode.put("mchid", wxPayProperties.getMchId())
                .put("appid", wxPayProperties.getAppId())
                .put("out_refund_no", out_refund_no)
                .put("notify_url", wxPayProperties.getRefundNotifyUrl())
                .put("out_trade_no", String.valueOf(orderId));
        rootNode.putObject("amount")
                .put("refund", order.getFee())
                .put("total", order.getFee())
                .put("currency", "CNY");
        String json = objectMapper.writeValueAsString(rootNode);
        log.info("退款接口请求参数: {}", json);
        objectMapper.writeValue(bos, rootNode);
        httpPost.setEntity(new StringEntity(bos.toString("UTF-8"), "UTF-8"));
        CloseableHttpResponse response = httpClient.execute(httpPost);
        String bodyAsString = EntityUtils.toString(response.getEntity());
        JSONObject jsonObject = JSONUtil.parseObj(bodyAsString);
        // 修改订单状态为退款中
        orderService.updateOrderStatus(orderId, OrderStatus.REFUNDING);

    }

    @Override
    public Boolean notifyRefund(HttpServletRequest request) {
        try {
            String requestBodyData = this.getRequestBodyData(request);
            log.info("接收微信退款接口回调请求, 请求参数: {}", requestBodyData);
            JSONObject jsonObject = JSONUtil.parseObj(requestBodyData);
            JSONObject resource = jsonObject.getJSONObject("resource");
            String plainText = this.decrypt(resource);
            log.info("解密后的明文信息为: {}", plainText);
            JSONObject obj = JSONUtil.parseObj(plainText);
            processRefund(jsonObject);
            return null;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    @Override
    public void processRefund(JSONObject jsonObject) {
        Long out_trade_no = jsonObject.getLong("out_trade_no");
        RLock lock = redisson.getLock("refund-" + out_trade_no);
        lock.lock(10, TimeUnit.SECONDS);
        try {
            // 修改订单未退款成功
            orderService.updateOrderStatus(out_trade_no, OrderStatus.REFUNDED);
            // 保存流水信息
        } finally {
            lock.unlock();
        }
    }


    /**
     * 解密
     *
     * @param resource
     * @return
     * @throws GeneralSecurityException
     */
    private String decrypt(JSONObject resource) throws GeneralSecurityException {
        AesUtil aesUtil = new AesUtil(wxPayProperties.getApiV3Key().getBytes(StandardCharsets.UTF_8));
        String ciphertext = resource.getStr("ciphertext");
        String associated_data = resource.getStr("associated_data");
        String nonce = resource.getStr("nonce");
        // 对密文解密
        String plainText = aesUtil.decryptToString(associated_data.getBytes(StandardCharsets.UTF_8),
                nonce.getBytes(StandardCharsets.UTF_8),
                ciphertext);
        return plainText;
    }


    /**
     * 获取request中body数据
     *
     * @throws IOException
     */
    private String getRequestBodyData(HttpServletRequest request) throws IOException {
        BufferedReader bufferReader = new BufferedReader(request.getReader());
        StringBuilder sb = new StringBuilder();
        String line;
        while ((line = bufferReader.readLine()) != null) {
            sb.append(line);
        }
        bufferReader.close();
        return sb.toString();
    }

}